# -*- Coding: utf-8 -*-

"""
iSuperpose

superimposition module
"""

import os.path
import subprocess
import copy
import shutil

from libiSuperpose.config import *
from libiSuperpose.io import *
from libiSuperpose.misc import *
from libiSuperpose.geom import *

BINPATH = os.path.dirname(os.path.abspath(__file__))

# -- superposition --
def onePDBxtractCrds(PDB, seqMask, atms, verbose = 0, strict = False, debug = 0):
    """
    Extract coordinates of a PDB instance according
    to the sequence mask, and atoms types.
    * JM 2008, May 29 
      PDB.aaseq().replace("X","") -> skip hetero for superposition
    * JM 2008, Sept 4
      This task should have been treated by the clean() method while
      launching PDB instance.
    """
    import PyPDB.PyPDB as PyPDB
    PDB = PyPDB.PDB(PDB, hetSkip = 1)
    if verbose:
        print >> sys.stdout, "onePDBxtractCrds: will proceed ... ", "strict", strict

    selCrds = []

    # if seqMask is empty consider all residues
    if seqMask == None:
        seqMask = [1] * len(aaseq2(PDB,atms).replace("X",""))
        
    if (len(aaseq2(PDB,atms).replace("X","")) != len(seqMask)):
        print >> sys.stderr, "[ERROR] onePDBxtractCrds: mask and PDB length differs (HET) (%d vs %d) for %s!"%(len(PDB),len(seqMask),PDB.id)
        # print seqMask, len(seqMask)
        # print aaseq2(PDB,atms), len(aaseq2(PDB,atms))
        # print strict,  strict and (len(aaseq2(PDB,atms).replace("X","")) != len(seqMask))
        sys.exit(0)

    # Do we have only hetero ?
    if len(seqMask) == 0:
        seqMask = [1]

    # for each residu of the PDB
    for aRes in range(0, len(seqMask), 1):

        if not seqMask[aRes]:
            continue
        else:
            if debug:
                print >> sys.stderr, "[REMARK] onePDBxtractCrds selected:",
                print >> sys.stderr, PDB[aRes].select(['CA']),
        # for each atom type
        # "CA" for example
        for aAtm in atms:
            try:
                aSel = PDB[aRes].select([aAtm])
            except:
                print >> sys.stderr, "[REMARK] onePDBxtractCrds: %s residu type not found !"%aAtm
                sys.exit(0)
            
            # JM 2008, oct 10
            # Do not consider residue without superposed atoms
            if not len(aSel):
                continue
                
            # keep only coordinates
            selCrds.append(aSel.xyz())
            #selCrds += aSel.xyz()

    return selCrds

def bestFitCrds(refCrds, mobiCrds, wPath = "./", verbose = 0, debug = 0):
    """
    Use QBestFit to superpose selected coordinates.
    Returns the transformation matrix.
    """

    if verbose:
        print >> sys.stdout, "bestFitCrds: will proceed ... "

    transMat = None

    if len(refCrds) != len(mobiCrds) :
        print >> sys.stderr, "[ERROR] bestFitCrds: Size of selected coordinates differs (%d VS %d) !"%(len(refCrds),len(mobiCrds))
        sys.exit(0)

    # Open the pipe to QBestFit
    cmd = "%s/QBestFit"%(BINPATH)
    p = subprocess.Popen( cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE, close_fds=True )
    (QBestFitF, QBestFitOF, QBestFitEF) = (p.stdin, p.stdout, p.stderr)

    # The number of coordinates to fit
    nCrds = len(refCrds)
    QBestFitF.write("%d\n"%(nCrds))
    
    if debug:
        print >> sys.stderr, "%d" % nCrds
        print >> sys.stderr, "ref:", refCrds
        print >> sys.stderr, "mobi:", mobiCrds
    
    # Write coordinates to fit
    printFloatList2(QBestFitF, refCrds)
    printFloatList2(QBestFitF, mobiCrds)
    QBestFitF.close()
    
    # Parse the temporary results file
    tMat, caRMSd = getQBestFitTransMat(QBestFitOF)
    
    return tMat, caRMSd

def superposedPDBsMeasures(refCrds, mobiCrds, wPath = "./", realPipe = 0, verbose = 0):
    """
    Computes:
      - cRMSd
      - GDT_TS
    On superposed structures.
    """

    if verbose:
        print >> sys.stdout, "superposedPDBsMeasures: will proceed ... "

    if len(refCrds) != len(mobiCrds) :
        print >> sys.stderr, "[ERROR] superposedPDBsMeasures: Size of selected coordinates differs (%d VS %d) !"%(len(refCrds),len(mobiCrds))
        sys.exit(0)

    if realPipe:

        # In this case, we write in stdin

        cmd = "%s/GDT_TS"%(BINPATH)

        p = subprocess.Popen( cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE, close_fds=True )
        (QBestFitF, QBestFitOF, QBestFitEF) = (p.stdin, p.stdout, p.stderr)

        # The number of coordinates to fit
        nCrds = len(refCrds)
        QBestFitF.write("%d\n"%(nCrds))
        # Write coordinates to fit
        printFloatList2(QBestFitF, refCrds)
        printFloatList2(QBestFitF, mobiCrds)
        QBestFitF.close()
        
    else:

        # Here, we write into a file

        tmpFN = "/tmp/iSuperpose-%d" % random.randint(0,100000000)

        try:
            QBestFitF = open(tmpFN,"w")
        except:
            print >> sys.stderr, "[ERROR] Could not open file %s !"%tmpFN
            sys.exit(0)

        # The number of coordinates to fit
        nCrds = len(refCrds)
        QBestFitF.write("%d\n"%(nCrds))
        # Write coordinates to fit
        printFloatList2(QBestFitF, refCrds)
        printFloatList2(QBestFitF, mobiCrds)
        QBestFitF.close() 

        cmd = "cat %s | %s/GDT_TS"%(tmpFN, BINPATH)

        p = subprocess.Popen( cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True )
        (QBestFitOF, QBestFitEF) = (p.stdout, p.stderr)


    oL = []
    while True:
        line = QBestFitOF.readline()        
        if not line:
            break
        oL.append(line)
    # oStr = p.communicate()[0]
    # oL = oStr.split("\n")

    cRMSd = float(oL[0].split()[1])
    GDT_TS = float(oL[1].split()[1])
    
    return cRMSd, GDT_TS


def onePDBTrans( PDB, tMat, verbose ):
    """
    Apply transformation matrix to the PDB instance.
    """

    if verbose:
        print >> sys.stdout, "onePDBTrans: will proceed ..."    

    PDB_trans = copy.deepcopy( PDB )

    # for each residu
    for aRes in PDB_trans:
        # for each atom
        for aAtm in aRes:

            # get coordinates to transform
            tCrds = crdRotate(aAtm.xyz(), tMat)

            # Mutate coordinates
            aAtm.txt = aAtm.txt[:30] + "%8.3f%8.3f%8.3f"%(tCrds[0],tCrds[1],tCrds[2]) + aAtm.txt[54:]

    return PDB_trans

def twoPDBspp(PDB_Ref, PDB_Mobi, seqMaskRef, seqMaskMobi, atms, wPath = "./", PDB_TM = None, verbose = 0):
    """
    Superpose 2 PDB files according to an alignment.
    PDBref      : reference PDB instance
    PDBmobi     : mobile PDB instance
    seqMaskRef  : sequences masks (residus to superpose) for reference
    seqMaskMobi : sequences masks (residus to superpose) for mobile    
    atms        : atoms list used to superpose
    PDB_TM      : apply transformation matrix to the additional PDB file
    verbose     : verbose mode
    """
    if verbose:
        print >> sys.stdout, "twoPDBspp: will proceed ..."
        
    # Extract coordinates according to the sequence mask
    # and atoms types
    PDB_Ref_Crds  = onePDBxtractCrds(PDB_Ref,  seqMaskRef, atms, verbose = verbose)
    PDB_Mobi_Crds = onePDBxtractCrds(PDB_Mobi, seqMaskMobi, atms, verbose = verbose)

    # print PDB_Ref_Crds
    # print seqMaskRef
    # print len(PDB_Ref_Crds)
    # print PDB_Mobi_Crds
    # print seqMaskMobi
    # print len(seqMaskMobi)
    # Write out used coordinates and use QBestFit to 
    # get the rotation matrix
    tMat, caRSMd_zone  = bestFitCrds(PDB_Ref_Crds, PDB_Mobi_Crds, wPath = wPath, verbose = verbose)
    if verbose:
        outMat_simple( tMat, hcomm = "# Transformation matrix", verbose = verbose )
    
    # Apply transformation matrix to the mobile PDB
    # instance
    #PDB_MobiTrans = onePDBTrans(PDB_Mobi, tMat, verbose)
    if PDB_TM:
        PDB_MobiTrans = onePDBTrans(PDB_TM, tMat, verbose)        
    else:
        PDB_MobiTrans = onePDBTrans(PDB_Mobi, tMat, verbose)

    return PDB_MobiTrans, caRSMd_zone

def twoPDBCompare(PDB_Ref, PDB_Mobi, atms, wPath = "./", verbose = 0):
    """
    Compute CA RMSd on 2 PDB files of the same length
    PDB structures are supposed to be superposed.
    """
    if verbose:
        print >> sys.stdout, "twoPDBCompare: will proceed ..."
        
    # Extract coordinates according to the sequence mask
    # and atoms types
    PDB_Ref_Crds  = onePDBxtractCrds(PDB_Ref,  None, atms, verbose)
    PDB_Mobi_Crds = onePDBxtractCrds(PDB_Mobi, None, atms, verbose)

    # Write out used coordinates and use QBestFit to 
    # get the rotation matrix
    cRSMd, GDT_TS = superposedPDBsMeasures(PDB_Ref_Crds, PDB_Mobi_Crds, wPath = wPath, verbose = verbose)

    return cRSMd, GDT_TS

def twoPDBTMalign( PDBrefFN, PDBmobiFN, wPath = "./", verbose = 0 ):
    """
    Performs a TM alignement of two PDB files.
    Return the parsed output in a TM object.
    (see parseTMalign for details).

    NB: required TMalign properly installed
    availability @
    http://zhang.bioinformatics.ku.edu/TM-align/TMalign.f
    """

    if verbose:
        print >> sys.stdout, "twoPDBTMalign: will TMalign %s (reference) and %s file (mobile) ..."%(PDBrefFN,PDBmobiFN)
        
    # temporary file to write in
    outTmpFN = "%s/tmp_TMalign_%.2f_%d"%(wPath, time.time(),random.randint(0,10000))

    # The system command
    cmd = "%s/TMalign %s %s > %s"%(BINPATH, PDBrefFN, PDBmobiFN, outTmpFN)
    
    # Launch the TMalignement
    try:
        os.system(cmd)
    except:
        print >> sys.stderr, "[ERROR] An error occured when launching TMalign on %s and %s files !"%(PDBrefFN, PDBmobiFN)
        sys.exit(0)

    TM = parseTMalign( outTmpFN )

    # remove temporary file
    try:
        os.remove( outTmpFN )
    except:
        print >> sys.stderr, "[ERROR] Impossible to remove %s temporary file !"%outTmpFN
        sys.exit(0)

    return TM

def twoPDBTMalign2( PDBrefFN, PDBmobiFN, wPath = "/tmp/", verbose = 0 ):
    """
    Performs a TM alignement of two PDB files.
    Return the parsed output in a TM object.
    (see parseTMalign for details).

    NB: required TMalign properly installed
    availability @
    http://zhang.bioinformatics.ku.edu/TM-align/TMalign.f

    2008, June 25:
    Long file names is no more a problem, we work in wPath.
    2008, Sept 8: ** BugFix **
    Now it's true :)
    
    """

    if verbose:
        print >> sys.stdout, "twoPDBTMalign2: will TMalign %s (reference) and %s file (mobile) ..."%(PDBrefFN,PDBmobiFN)

    useRelPath = 0

    if useRelPath:
        # Current directory
        oriPath = os.getcwd()
        PDBrefFNRel = os.path.basename( PDBrefFN )
        PDBmobiFNRel = os.path.basename( PDBmobiFN )

        # Go to working directory
        os.chdir( wPath )
        # link the reference structure
        if not os.path.exists( "./%s" % PDBrefFNRel ):
            os.symlink( PDBrefFN, PDBrefFNRel )
        # link the mobile structure
        if not os.path.exists( "./%s" % PDBmobiFNRel ):
            os.symlink( PDBmobiFN, PDBmobiFNRel )
        ref = PDBrefFNRel
        mobi = PDBmobiFNRel
    else:
        # Here we hard copy files to wPath
        tmpRefFN  = "%s/%s" % (wPath,os.path.basename(PDBrefFN))
        tmpMobiFN = "%s/%s" % (wPath,os.path.basename(PDBmobiFN))
        if not os.path.exists(tmpRefFN):
            shutil.copy( PDBrefFN,  tmpRefFN )
        if not os.path.exists(tmpMobiFN):
            shutil.copy( PDBmobiFN, tmpMobiFN )        
        ref  = tmpRefFN
        mobi = tmpMobiFN

    # The system command
    cmd = "%s/TMalign %s %s "%( BINPATH, ref, mobi )

    p = subprocess.Popen( cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE, close_fds=True )
    (child_stdin, child_stdout, child_stderr) = (p.stdin, p.stdout, p.stderr)

    TM = parseTMalign2( child_stdout, verbose = verbose )

    # Restore path
    if useRelPath:
        if os.path.exists( PDBrefFNRel ):
            os.unlink( PDBrefFNRel )
        if os.path.exists( PDBmobiFNRel ):
            os.unlink( PDBmobiFNRel )
        os.chdir( oriPath )

    return TM

def TM2seqMasks( TM, PDBrefSeq, PDBmobiSeq, verbose = 0, debug = 0 ):
    """
    Generate PDB files sequences masks from TMalign output.
    This mask will tell wich residues are selected for superposition.
    """
    
    refMask  = [0] * len(PDBrefSeq)
    mobiMask = [0] * len(PDBmobiSeq)

    aRefP  = 0
    aMobiP = 0

    # for each position of the alignement
    for i in range(0, TM['maxAlnL'], 1):

        if TM['alignId'][i] == ":" or TM['alignId'][i] == ".":

            if TM['alignRef'][i] == PDBrefSeq[aRefP]:
                refMask[aRefP] = 1
                
            if TM['alignMobi'][i] == PDBmobiSeq[aMobiP]:
                mobiMask[aMobiP] = 1        

        if i < len( TM['alignRef'] ) and  TM['alignRef'][i] != '-' and TM['alignRef'][i] != ' ':
            aRefP  += 1
        
        if i < len( TM['alignMobi'] ) and  TM['alignMobi'][i] != '-' and TM['alignMobi'][i] != ' ':
            aMobiP += 1
    
    if debug:
        print TM['alignRef']
        print PDBrefSeq
        print str(refMask).replace(' ','').replace('[','').replace(',','').replace(']','')
        print TM['alignMobi']
        print PDBmobiSeq
        print str(mobiMask).replace(' ','').replace('[','').replace(',','').replace(']','')
    
    return (refMask, mobiMask)

