# -*- Coding: utf-8 -*-

"""
iSuperpose

I/O module
"""

import random
import math
import os.path
import string
import time
import sys

import PyPDB.PyPDB as PDB

def loadPDBfile( PDBFN, hetSkip = 0, verbose = 0):
    """
    Load a PDB file according to PDBPy module.    
    Return: the PDB instance or None if empty
    * JM 2008, May 29 
      Hetero are not skiped (were lost in the Mobile structure)
    """

    PDBF = None

    if verbose:
        print >> sys.stdout, "loadPDBfile: Will load %s PDB file ... "%PDBFN,

    try:
        PDBF = PDB.PDB(PDBFN, hetSkip = hetSkip, altCare = 1)
        PDBF.clean()
        if verbose:
            print >> sys.stdout, "ok"
    except:
        print >> sys.stderr, "[ERROR] while loading %s PDB file !"%PDBFN
        sys.exit(0)
        
    return PDBF

def outputPDB( outPDBFN, PDBfitted, CAonly=False, verbose = 0 ):

    if verbose:
        print >> sys.stdout, "outputPDB: Will write PDB instance to file %s... "%outPDBFN,

    if CAonly:
        PDBfitted = PDBfitted.select(awhat=['CA'])

    if outPDBFN == None:
        PDBfitted.out()
    else:
        try:
            PDBfitted.out(outName = outPDBFN)
        except:
            print >> sys.stderr, "[ERROR] Impossible to write in %s ouput file, will use stdout."%outPDBFN
            PDBfitted.out()
            
    return

def fileGetLines(FN, verbose = 0):
    """
    Open txt file, read and returns lines.
    """

    if verbose:
        print >> sys.stdout, "fileGetLines: Will read %s file lines ... "%FN

    lines = []    

    try:
        F = open(FN,"r")
    except:
        print >> sys.stderr, "[ERROR] Could not open file %s !"%FN
        sys.exit(0)

    try:
        lines = F.readlines()
    except:
        print >> sys.stderr, "[ERROR] Could not parse file %s !"%FN
        sys.exit(0)

    try:
        F.close()
    except:
        print >> sys.stderr, "[ERROR] Could not close file %s !"%FN
        sys.exit(0)

    return lines


# -- alignments --
def parseTMalign( alnFN, verbose = 0 ):
    """
    Parse TM align file format:

 **************************************************************************
 *                               TM-align                                 *
 * A protein structural alignment algorithm based on TM-score             *
 * Reference: Y. Zhang and J. Skolnick, Nucl. Acids Res. 2005 33, 2302-9  *
 * Comments on the program, please email to: zhang6@buffalo.edu           *
 **************************************************************************

Chain 1:1R69.pdb    Size=  63
Chain 2:2CRO.pdb    Size=  65 (TM-score is normalized by   65)

Aligned length=  63, RMSD=  0.79, TM-score=0.89969, ID=0.524

(":" denotes the residue pairs of distance < 5.0 Angstrom)
--SISSRVKSKRIQLGLNQAELAQKVGTTQQSIEQLENGKTKRPRFLPELASALGVSVDWLLNGT
  :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
MQTLSERLKKRRIALKMTQTELATKAGVKQQSIQLIEAGVTKRPRFLFEIAMALNCDPVWLQYGT

    --
    And returns a TM align dictionnary with the following keys:
    
      refId     : reference PDB file Id
      mobiId    : mobile PDB file Id
      refSze    : reference PDB file size
      mobiSze   : mobile PDB file size
      alignL    : alignement length
      rmsd      : CA rmsd of aligned regions
      TMScore   : the score
      ID        : ID
      alignRef  : the alignement line for reference PDB
      alignId   : the alignement line for identities
      alignMobi : the alignement line for mobile PDB
      maxAlnL   : maximal alignement line length
    """

    if verbose:
        print >> sys.stdout, "parseTMalign: will parse %s file ... "%alnFN

    TM = {}
    
    lines  = fileGetLines(alnFN)
    nLines = len(lines)

    if nLines == 0:
        print >> sys.stderr, "[ERROR] parseTMalign: %s file contains no lines !"%alnFN
        sys.exit(0)
    
    for aL in range(0, nLines, 1):

        line = lines[aL]

        # skip commentary and empty lines
        if len(line) < 2  or line[1] == "*" or line[0] == "(":
            continue

        if   line[:7] == "Chain 1":
            TM['refId']  = line[8:18]
            TM['refSze'] = int( line[25:29] )
            
        elif line[:7] == "Chain 2":
            TM['mobiId']  = line[8:18]
            TM['mobiSze'] = int( line[25:29] )
                    
        elif line[:7] == "Aligned":
            TM['alignL']  = int(   line[15:19] )
            TM['rmsd']    = float( line[26:32] )
            TM['TMScore'] = float( line[43:50] )
            TM['ID']      = float( line[55:-1] )

        # the last 3 lines present the alignement
        elif aL == nLines -4:
            TM['alignRef']  = line[:-1]
        elif aL == nLines -3:
            TM['alignId']  = line[:-1]
        elif aL == nLines -2:
            TM['alignMobi']  = line[:-1]

    # the maximal alignement line length
    if len(TM['alignRef']) >= len(TM['alignMobi']):
        TM['maxAlnL'] = len(TM['alignRef'])
    else:
        TM['maxAlnL'] = len(TM['alignMobi'])

    return TM

def parseTMalign2( alnF, verbose = 0 ):
    """
    Parse TM align file format:

 **************************************************************************
 *                               TM-align                                 *
 * A protein structural alignment algorithm based on TM-score             *
 * Reference: Y. Zhang and J. Skolnick, Nucl. Acids Res. 2005 33, 2302-9  *
 * Comments on the program, please email to: zhang6@buffalo.edu           *
 **************************************************************************

Chain 1:1R69.pdb    Size=  63
Chain 2:2CRO.pdb    Size=  65 (TM-score is normalized by   65)

Aligned length=  63, RMSD=  0.79, TM-score=0.89969, ID=0.524

(":" denotes the residue pairs of distance < 5.0 Angstrom)
--SISSRVKSKRIQLGLNQAELAQKVGTTQQSIEQLENGKTKRPRFLPELASALGVSVDWLLNGT
  :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
MQTLSERLKKRRIALKMTQTELATKAGVKQQSIQLIEAGVTKRPRFLFEIAMALNCDPVWLQYGT

    --
    And returns a TM align dictionnary with the following keys:
    
      refId     : reference PDB file Id
      mobiId    : mobile PDB file Id
      refSze    : reference PDB file size
      mobiSze   : mobile PDB file size
      alignL    : alignement length
      rmsd      : CA rmsd of aligned regions
      TMScore   : the score
      ID        : ID
      alignRef  : the alignement line for reference PDB
      alignId   : the alignement line for identities
      alignMobi : the alignement line for mobile PDB
      maxAlnL   : maximal alignement line length
    """

    if verbose:
        print >> sys.stderr, "parseTMalign2: will parse file ... "

    TM = {}
    
    lines  = alnF.readlines()
    nLines = len(lines)

    if verbose:
        for line in lines:
            print >> sys.stderr, line,

    if nLines == 0:
        print >> sys.stderr, "[ERROR] parseTMalign2: file contains no lines !"
        sys.exit(0)
    
    for aL in range(0, nLines, 1):

        line = lines[aL]

        # skip commentary and empty lines
        if len(line) < 2  or line[1] == "*" or line[0] == "(":
            continue

        if   line[:7] == "Chain 1":
            TM['refId']  = line[8:18]
            TM['refSze'] = int( line[25:29] )
            
        elif line[:7] == "Chain 2":
            TM['mobiId']  = line[8:18]
            TM['mobiSze'] = int( line[25:29] )
                    
        elif line[:7] == "Aligned":
            TM['alignL']  = int(   line[15:19] )
            TM['rmsd']    = float( line[26:32] )
            TM['TMScore'] = float( line[43:50] )
            TM['ID']      = float( line[55:-1] )

        # the last 3 lines present the alignement
        elif aL == nLines -4:
            TM['alignRef']  = line[:-1]
        elif aL == nLines -3:
            TM['alignId']  = line[:-1]
        elif aL == nLines -2:
            TM['alignMobi']  = line[:-1]

    # the maximal alignement line length
    if len(TM['alignRef']) >= len(TM['alignMobi']):
        TM['maxAlnL'] = len(TM['alignRef'])
    else:
        TM['maxAlnL'] = len(TM['alignMobi'])

    return TM


def TMaln2fastaOut( TM, refFN, mobiFN, outFN, ids = 0):

    maxChar = 80

    if outFN != None:
        stream = open(outFN,"a")
    else:
        stream = sys.stdout

    stream.write("\n# TM-alignment\n")
    
    nChar = 0

    if ids:
        fFN = ids[0]
    else:
        fFN = string.replace(refFN,"/data/www/htdocs/tmp/iSuperpose/","")
        
    # stream.write(">%s (%d aa)\n"%(refFN,TM['refSze']))
    stream.write(">%s (%d aa)\n"%(fFN,TM['refSze']))
    
    for aChar in range(0,len(TM['alignRef']),1):

        if( nChar == maxChar ):
                stream.write("\n")
                nChar = 0
                
        # if there is an identity
        if TM['alignId'][aChar] == ":" or TM['alignId'][aChar] == ".":
            stream.write(string.upper(TM['alignRef'][aChar]))
        else:
            stream.write(string.lower(TM['alignRef'][aChar]))

        nChar += 1
        
    stream.write("\n")
    nChar = 0
    
    if ids:
        iFN = ids[1]
    else:
        iFN = string.replace(mobiFN,"/data/www/htdocs/tmp/iSuperpose/","")

    # stream.write(">%s (%d aa) TM-Score:%.3f with ref %s\n"%(mobiFN,TM['mobiSze'],TM['TMScore'],refFN))
    stream.write(">%s (%d aa) TM-Score:%.3f with ref %s\n"%(iFN,TM['mobiSze'],TM['TMScore'],fFN))
    
    for aChar in range(0,len(TM['alignMobi']),1):
        
        if( nChar == maxChar ):
                stream.write("\n")
                nChar = 0
                
        # if there is an identity
        if TM['alignId'][aChar] == ":" or TM['alignId'][aChar] == ".":
            stream.write(string.upper(TM['alignMobi'][aChar]))
        else:
            stream.write(string.lower(TM['alignMobi'][aChar]))

        nChar += 1
        
    stream.write("\n")
    
    if outFN != None:
        stream.close()
        
    return 

def TMmatOut( TMmat, ids = 0, outFN = None ):
    """
    Output the TM-align Scores Matrix
    """

    if outFN != None:
        stream = open(outFN,"w")
    else:
        stream = sys.stdout

    stream.write("# TMScores Matrix generated by iSuperpose\n")
    
    if ids:
        stream.write("#Ref/Mob ")
        for aId in ids[1]:
            stream.write("%8s "%aId)
        stream.write("\n")
    
    #for aC in range(0,len(TMmat['mobiles']),1):
    #    stream.write("%5d "%aC)
        
    for aL in range(0,len(TMmat['references']),1):
        if ids:
            stream.write("%8s "%ids[0][aL])
        for aC in range(0,len(TMmat['mobiles']),1):
            stream.write("%8.3f "%TMmat['scores'][aL][aC])
        stream.write("\n")
    
    if outFN != None:
        stream.close()

    return

def parseSimpleAlign( alnFN, PDBrefSeq, PDBmobiSeq, verbose = 0 ):
    """
    Parse a simple alignment text file with format like:
    
# 24/04/06 test alignement manuel rebuilt/reference
# pour la cible CASP6 T0227
reference MERPKLGLIVREPYASLIVDGRKVWEIRRRKTRHRGPLGIVSGGRLIG
                     ||||||||||||||||||
rebuilt   MERPKLGLIVREPYASLIVDGRKVWEIRRRKTRHRGPLGIVSGGRLIG

reference QADLVGV------------------------------PLYAWVLENAF

rebuilt   QADLVGVEGPFSVEELLAHQEKHLAEEAFLRAYAKDEPLYAWVLENAF

reference RYEKPLHVPMFVDLSEVR

rebuilt   RYEKPLHVPRRPG
    """

    if verbose:
        print >> sys.stdout, "parseSimpleAlign: will parse %s file ..."%alnFN

    # The identity special character
    idChar = "|"
    
    refMask  = [0] * len(PDBrefSeq)
    mobiMask = [0] * len(PDBmobiSeq)

    refPos  = 0
    mobiPos = 0

    lines  = fileGetLines(alnFN)
    nLines = len(lines)

    if nLines == 0:
        print >> sys.stderr, "[ERROR] parseSimpleAlign: %s file contains no lines !"%alnFN
        sys.exit(0)

    aL = 0
    
    while aL < nLines-2:
        
        lineRef  = lines[aL]
        lineId   = lines[aL+1]
        lineMobi = lines[aL+2]

        # skip commentary and empty lines
        if "#" in lineRef or len(lineRef) < 2 :
            aL += 1
            continue

        # split the reference line
        splitedRef = lineRef.split()
        if len(splitedRef) != 2:
            print >> sys.stderr, "[ERROR] parseSimpleAlign: bad alignment format in %s file !"%alnFN
            print >> sys.stderr, "        line %d: %s"%(aL+1,lineRef)
            sys.exit(0)

        # index of the 1 residue of the sequence
        startPos = lineRef.index(splitedRef[1])

        # for each char:
        for aPos in range(startPos,len(lineId),1):

            # We found an identity
            if lineId[aPos] == idChar:
                refMask[refPos] += 1
                mobiMask[mobiPos] += 1

            if lineRef[aPos] != "-" and lineRef[aPos] != " " and lineRef[aPos] != "\n" and lineRef[aPos] == PDBrefSeq[refPos]:
                refPos += 1
            if lineMobi[aPos] != "-" and lineMobi[aPos] != " " and lineMobi[aPos] != "\n" and lineMobi[aPos] == PDBmobiSeq[mobiPos]:
                mobiPos += 1
                
        aL += 3

    # check alignement length and PDB sequence
    if refPos != len(PDBrefSeq):
        print >> sys.stderr, "[ERROR] Sequences lengths for alignement and PDB file does not match for reference structure (%d VS %d)"%(refPos, len(PDBrefSeq))
        sys.exit(0)
        
    if mobiPos != len(PDBmobiSeq):
        print >> sys.stderr, "[ERROR] Sequences lengths for alignement and PDB file does not match for mobile structure (%d VS %d)"%(mobiPos, len(PDBmobiSeq))
        sys.exit(0)

    return (refMask, mobiMask)

def fasta2SeqMask( seq, debug = 0 ):
    """    
    Parse fasta sequence format into 3Dmask:
    
      * UPPER CASE: ignore this residu for superposition 
      * lower case: consider this residu in superposition
      * gaps are ignored ('_' or '-' or ' ')

    """

    mask = []

    for aChar in seq:

        # skip gaps
        if aChar == "" or aChar == " " or aChar == "-" or aChar == "_":
            continue

        # use this residu for superposition
        if aChar.isupper():
            mask.append(1)
        # don't use this one
        else:
            mask.append(0)
    if debug:
        print >> sys.stderr, "fasta2SeqMask:",mask
            
    return mask

def fasta2SeqMask2( refSeq, mobiSeq, debug = 0 ):
    """    
    Parse fasta sequence format into 3Dmask:
    
      * UPPER CASE: ignore this residu for superposition 
      * lower case: consider this residu in superposition
      * gaps are ignored ('_' or '-' or ' ')

    aSeqMaskRef  = fasta2SeqMask2( refSeq[1], mobiSeqs[aMobiRk][1] )
    
    """

    mask = []

    for aChar in mobiSeq:

        # skip gaps
        if aChar == "" or aChar == " " or aChar == "-" or aChar == "_":
            mask.append(0)
            continue

        # use this residu for superposition
        if aChar.isupper():
            mask.append(1)
        # don't use this one
        else:
            mask.append(0)
    if debug:
        print >> sys.stderr, "fasta2SeqMask:",mask
            
    return mask

def parseFastaAlign( alnFN, verbose = 0, debug = 0):
    """
    Parse fasta format:

      * first sequence must be the reference
      
      Example:
      
      > reference
      ATRTTRTAAAaaayippp...
      > mobile1
      aRFTVNCFFFRTTilipp...
      > mobile2    
      ATRYePeIePlpplllpp...
      
    """
    
    if verbose:
        print >> sys.stdout, "parseFastaAlign: will parse %s file ..."%alnFN

    lines  = fileGetLines(alnFN)
    nLines = len(lines)
    if nLines == 0:
        print >> sys.stderr, "[ERROR] parseFastaAlign: %s file contains no lines !"%alnFN
        sys.exit(0)

    seqs = []
    masks = []
    
    for aL in range(0,len(lines),1):

        #  skip commentary and empty lines
        if lines[aL][0] == "#" or lines[aL] == "" or lines[aL] == "\n" or lines[aL] == " ":
            continue

        # new sequence
        if lines[aL][0] == ">":
            tmpSeq = ""
            aId = lines[aL][:-1]
            continue
        
        tmpSeq += lines[aL][:-1]

        # last sequence or a new one is comming
        # store sequence
        if aL == len(lines) -1 or lines[aL+1][0] == ">":
            seqs.append([aId,tmpSeq])

    if debug:
        print >> sys.stderr, "parseFastaAlign:"
        for i in seqs:
             print >> sys.stderr, "%s"%i[0]
             print >> sys.stderr, "%s"%i[1]
        sys.exit(0)
        
    return seqs

def getQBestFitTransMat(iF, verbose = 0):
    """
    Parse QBestFit output transformation matrix file and returns the matrix.
    """

    if verbose:
        print >> sys.stdout, "getQBestFitTransMat: will parse %s file ... "%FN

    tMat = []

    tmpLines = iF.readlines()

    # Get CA RMSd
    try:
        caRMSd = float(tmpLines[1].split()[1])
    except:
        if verbose:
            print >> sys.stderr, "[ERROR] getQBestFitTransMat: Could not cast RMSd !"
        caRMSd = 0.

    # parse tmp file
    for line in tmpLines[2:6]:
        splited = line.split()
        tmp = []
        for i in splited:
            try:
                tmp.append(float(i))                
            except:
                print >> sys.stderr, "[ERROR] getQBestFitTransMat: Could not cast type !"                
                sys.exit(0)
        tMat.append(tmp)

    return tMat, caRMSd

# -- PDB extracting functions --
def PDBZones( iPDB, zones = None, verbose = 0 ):
    """
    Extract a PDB giving zones definition
    """
    
    c = 0
    if zones:
        for aZone in zones:
            # negative index compatibility
            if aZone[0] < 0:
                aZone[0] = len(iPDB) + aZone[0]
            if aZone[1] < 0:
                aZone[1] = len(iPDB) + aZone[1]

            theSze = int( math.fabs( aZone[1] - aZone[0] ) ) +1
            if c == 0:
                oPDB = iPDB.mask( aZone[0], aZone[1]+1, mask = "X" * theSze )
            else:
                oPDB += iPDB.mask( aZone[0], aZone[1]+1, mask = "X" * theSze )
            c += 1
    else:
        oPDB = iPDB

    return oPDB

# MODELS
def extractOnePDBModel( PDB, modelNum, FN, zones = None, verbose = 0 ):
    """
    Extract one model from multiple models PDB file
    to disk with file name FN.
    """
    
    nModels = PDB.nModels()

    # if the model exists 
    if nModels >= modelNum :
        try:
            # launch the model
            if PDB.curMdl != modelNum:
                PDB.setModel(modelNum)
            # Force output of the file even if it's a simple one
            # to take care of hetero, alternates ...
            PDB.clean()
            PDB.id  = FN.split("/")[-1] 

            if zones:
                PDB = PDBZones( PDB, zones = zones, verbose = 0 )
            
            # ouput file to disk
            PDB.out(FN)
        except:
            print >> sys.stderr, "[ERROR] Impossible to extract the model '%d' (max: %d)"%(modelNum,nModels)
            return 0
    else:
        print >> sys.stderr, "[ERROR] The model '%d' does not exists (max: %d)."%(modelNum,nModels)
        return 0

    return 1

def extractPDBModels( PDB, FNRoot, zones = None, verbose = 0 ):
    """
    Extract all the models of a multiple models PDB
    file with file name FNRoot + Model number.
    Returns output file names list.
    """
    nModels = PDB.nModels()
    fileNames = []
    
    for aModel in range(1,nModels+1,1):
        if FNRoot[-4:] == ".pdb":
            aFN = FNRoot[:-4]
        else:
            aFN = FNRoot
        aFN += str(aModel) + ".pdb"
        if( not extractOnePDBModel( PDB, aModel, aFN, zones = zones ) ):
            sys.exit(0)
        fileNames.append(aFN)
        
    return fileNames

# MULTIPLE PDB FILES
def extractOnePDBFromPDBList( PDBList, PDBNum, FN, zones = None, verbose = 0 ):
    """
    Extract one model from multiple models PDB file
    to disk with file name FN.
    """
    
    nPDB = len( PDBList )
    
    # if the model exists
    if nPDB > PDBNum:
        try:
            # Force output of the file even if it's a simple one
            # to take care of hetero, alternates ...
            aPDB = PDBList[PDBNum]
            aPDB.clean()
            aPDB.id  = FN.split("/")[-1] 

            if zones:
                aPDB = PDBZones( aPDB, zones = zones, verbose = 0 )

            # output file to disk
            aPDB.out(FN)
        except:
            print >> sys.stderr, "[ERROR] Impossible to extract the PDB '%d' from list (max: %d)"%(PDBNum,nPDB)
            return 0
    else:
        print >> sys.stderr, "[ERROR] The PDB '%d' from list does not exists (max: %d)."%(PDBNum,nPDB)
            
    return 1
                                                                                                            
def extractAllPDBFromPDBList( PDBList, FNRoot, zones = None, verbose = 0 ):
    """
    Extract all the models of a multiple PDB
    file with file name FNRoot + number.
    Returns output file names list.
    """
    nPDB = len( PDBList )
    fileNames = []
    
    for i in range(0,nPDB,1):
        if FNRoot[-4:] == ".pdb":
            aFN = FNRoot[:-4]
        else:
            aFN = FNRoot
        aFN += "%04d" %i + ".pdb"
        if( not extractOnePDBFromPDBList( PDBList, i, aFN, zones = zones) ):
            sys.exit(0)
        fileNames.append(aFN)
        
    return fileNames


# -- Advanced I/O --

def catSuperposedFiles( outFN, f2Cat, isModel, ids = 0, hetSkip = 0, CAonly = False, verbose = 0):

    if verbose:
        print >> sys.stderr, "\n>> catSuperposedFiles: will write to '%s' file ..."%outFN

    aModel = 1
    
    # if output file exists move
    if os.path.exists(outFN):        
        newFN  = outFN + ".%d"%time.time()
        os.rename(outFN,newFN)
        print >> sys.stderr, "Warning: the existing file '%s' has been moved to '%s' !"%(outFN,newFN)
        
    # for each file of the list
    for aFN in f2Cat:

        if verbose:
            print >> sys.stderr, "catSuperposedFiles: current file is %s" % aFN

        # open this file
        aPDB = loadPDBfile( aFN, hetSkip = hetSkip, verbose = verbose )
        
        if CAonly:
            aPDB = aPDB.select(awhat=['CA'])

        #if ids:
        #    aPDB.id = ids[aModel-1]

        # and output it
        if isModel:
            # first model
            if aModel == 1:
                aPDB.out( outName = outFN, fmode = "a", model = aModel )
            # last model
            elif aModel == len(f2Cat):
                aPDB.out( outName = outFN, fmode = "a", header = 0, model = aModel, end = 1 )
            # model
            else:
                aPDB.out( outName = outFN, fmode = "a", header = 0, model = aModel )
        else:
            aPDB.out( outName = outFN, fmode = "a", end = 1 )
            
        aModel += 1

    return 1


def eraseTempFiles( fList, verbose = 0 ):
    """
    Erase temporary files from disk (extracted models,
    or multiple PDB ... ).
    """
    for aFN in fList:
        if not aFN:
            continue
        if os.path.exists(aFN):
            try:
                os.remove(aFN)
            except:
                print >> sys.stderr, "[ERROR] Impossible to erase %s file from disk !"%(aFN)
                sys.exit(0)

    return 1

def outRMSdMat( mat, ids = 0, atms = "", hcomm = "", outFN = "", verbose = 0):

    if outFN != "":
        try:
            outF = open(outFN, "w")
        except:
            print >> sys.stderr, "[ERROR] outRMSdMat: Impossible to output %s file !"%outFN
            sys.exit(0)
    else:
        outF = sys.stdout
        
    n=len(atms)
    a = 1
    
    outF.write("# RMSd computed on ")
    for atm in atms:
        outF.write("%s"%atm)
        if a < n:
            outF.write(", ")
        else:
            outF.write("\n# %s\n"%hcomm)
        a += 1

    if ids:
        outF.write("#Ref/Mob ")
        for aId in ids[1]:
            outF.write("%8s "%aId)
    outF.write("\n")
    
    for aL in range(0,len(mat),1):
        if ids:
            outF.write("%8s "%ids[0][aL])
        for aC in range(0,len(mat[aL]),1):
            if mat[aL][aC] != "NA":
                outF.write("%8.3f "%mat[aL][aC])
            else:
                outF.write("%8s "%mat[aL][aC])
        outF.write("\n")
        
    if outFN != "":
        try:
            outF.close()
        except:
            print >> sys.stderr, "[ERROR] outRMSdMat: Impossible to close output file !"%outFN
            sys.exit(0)    
    
    return 1

def outMat( mat, ids = 0, hcomm = "", outFN = "", verbose = 0):

    if outFN != "":
        try:
            outF = open(outFN, "w")
        except:
            print >> sys.stderr, "[ERROR] outMat: Impossible to output %s file !"%outFN
            sys.exit(0)
    else:
        outF = sys.stdout
        
    outF.write( hcomm )

    if ids:
        outF.write("#Ref/Mob ")
        for aId in ids[1]:
            outF.write("%8s "%aId)
    outF.write("\n")
    
    for aL in range(0,len(mat),1):
        if ids:
            outF.write("%8s "%ids[0][aL])
        for aC in range(0,len(mat[aL]),1):
            if mat[aL][aC] != "NA":
                outF.write("%8.3f "%mat[aL][aC])
            else:
                outF.write("%8s "%mat[aL][aC])
        outF.write("\n")
        
    if outFN != "":
        try:
            outF.close()
        except:
            print >> sys.stderr, "[ERROR] outMat: Impossible to close output file !"%outFN
            sys.exit(0)    
    
    return 1

def outMat_simple( mat, ids = [], hcomm = "", outFN = "", verbose = 0):

    if outFN != "":
        try:
            outF = open(outFN, "w")
        except:
            print >> sys.stderr, "[ERROR] outMat_simple: Impossible to output %s file !"%outFN
            sys.exit(0)
    else:
        outF = sys.stdout
        
    outF.write( hcomm )

    if len(ids):
        outF.write("# ")
        for aId in ids[1]:
            outF.write("%8s "%aId)
    outF.write("\n")
    
    for aL in range(0,len(mat),1):
        if len(ids):
            outF.write("%8s "%ids[0][aL])
        for aC in range(0,len(mat[aL]),1):
            if mat[aL][aC] != "NA":
                outF.write("%8.3f "%mat[aL][aC])
            else:
                outF.write("%8s "%mat[aL][aC])
        outF.write("\n")
        
    if outFN != "":
        try:
            outF.close()
        except:
            print >> sys.stderr, "[ERROR] outMat_simple: Impossible to close output file !"%outFN
            sys.exit(0)    
    
    return 1


def extractPDB( PDBFN, zones = None, relative = 0, hetSkip = 1, verbose = 1 ):
    """
    This function extract PDB file from file name.
    Returns a list with the method at first rank
    followed by the extracted file names.
    Methods are:
             * 1 if simple PDB file
             * 2 if the PDB file is a PDBList
             * 3 if the PDB file contains models
             * 0 if empty 
    """

    if PDBFN[-4:] == ".pdb":
        xtractFNRoot = PDBFN[:-4]
    else:
        xtractFNRoot = PDBFN
    xtractFNRoot += "_xtract%d_"%random.randint(0,10000000)
    
    if relative:
        xtractFNRoot = xtractFNRoot.split("/")[-1]
    
    if verbose:
        print >> sys.stdout, "extractPDB: will process from '%s' file ..."%PDBFN

    # Is it a PDBList ?
    try:
        aPDBL = PDB.PDBList( PDBFN, hetSkip = hetSkip, altCare = 1 )
        PDBListLen = len(aPDBL)
        if PDBListLen == 0:
            if not os.path.exists( PDBFN ):                
                print >> sys.stderr, " Error: %s PDB file does not exists !" % PDBFN
            else:
                print >> sys.stderr, " Error: %s PDB file is empty !" % PDBFN
            sys.exit(0)
        elif PDBListLen > 1:
            if verbose:
                print >> sys.stdout, "Input PDB file '%s' is a PDB list (%d PDB), I will extract PDB files ..."%(PDBFN,len(aPDBL))
            outFiles = extractAllPDBFromPDBList( aPDBL, xtractFNRoot, zones = zones )
            del(aPDBL)
            return [2] + outFiles
        else:
            # Does this PDB file have models ?
            aPDB = aPDBL[0]
            nModels = aPDB.nModels()
            if nModels > 1:
                if verbose:
                    print >> sys.stdout, "Input PDB file '%s' has %d models, I will extract them into PDB files ..."%(PDBFN,nModels)
                outFiles = extractPDBModels( aPDB, xtractFNRoot, zones = zones  )
                del(aPDB)
                return [3] + outFiles
            else:
                # This is a simple PDB File
                if verbose:
                    print >> sys.stdout, "Input PDB file '%s' is a simple PDB file ..."%(PDBFN)
                # Force output of the file even if it's a simple one
                # to take care of hetero, alternates ...
                curOutFN = PDBFN.replace(".pdb","")+"-%d-iSuperpose.pdb" % random.randint(0,10000000)
                aPDB.clean()
                aPDB = PDBZones( aPDB, zones = zones, verbose = 0 )
                aPDB.id  = curOutFN.split("/")[-1]
                aPDB.out(outName = curOutFN)
                del(aPDB)
                return [1, curOutFN]
    except IOError:
        print >> sys.stderr, "[ERROR] Impossible to launch PDBList with %s file."%PDBFN

    return [0]
