import os
from setuptools import setup
from setuptools.command.install import install
from distutils.command.build import build
from distutils.errors import DistutilsSetupError
from subprocess import call

BASEPATH = os.path.dirname(os.path.abspath(__file__))
SRC_PATH = os.path.join(BASEPATH, 'src')
I_SUPERPOSE = "libiSuperpose"
class iSuperposeBuild(build):
    def run(self):
        # run original build code
        build.run(self)

        # build dependencies
        build_path = os.path.abspath(self.build_temp)

        cmd = [
            'make',
            'all',
            'OUT=' + build_path,
            'V=' + str(self.verbose),
        ]

        target_files = ["GDT_TS", "QBestFit", "TMalign"]

        def compile():
            call(["make", "clean"], cwd=SRC_PATH)
            call(cmd, cwd=SRC_PATH)

        self.execute(compile, [], 'Compiling dependencies')

        # copy resulting tool to library build folder
        self.mkpath(self.build_lib)

        if not self.dry_run:
            for target in target_files:
                targetfile = os.path.join(SRC_PATH, target)
                if not os.path.exists(targetfile):
                    raise DistutilsSetupError(
                      "Compilation of '%s' went wrong, file not present" % targetfile
                    )
                self.copy_file(targetfile, os.path.join(self.build_lib, I_SUPERPOSE))


class iSuperposeInstall(install):
    def initialize_options(self):
        install.initialize_options(self)
        self.build_scripts = None

    def finalize_options(self):
        install.finalize_options(self)
        self.set_undefined_options('build', ('build_scripts', 'build_scripts'))

    def run(self):
        # run original install code
        install.run(self)

        # install iSuperpose executables
        self.copy_tree(self.build_lib, self.install_lib)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "iSuperpose",
    #py_modules = ["libiSuperpose"],
    packages = [I_SUPERPOSE],
    install_requires = ['pyPDB>=10.0'],
    scripts = ["iSuperpose.py"],
    version = "1.9",
    description = "Protein structures superposition",
    author = "Julien Maupetit",
    author_email = "julien.maupetit@univ-paris-diderot.fr",
    url = "http://bioserv.rpbs.univ-paris-diderot.fr",
    keywords = ["protein", "PDB", "superimposition"],
    classifiers = [
        "Programming Language :: Python",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: Unix",
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    long_description = """\
iSuperpose : Yet Another Tool for Protein Structures Superimposition

Written (2004-2011) by J. Maupetit, Paris Diderot University

This tool is a wrapper that assembles existing tool and provides cool options that makes it a high value-added software. 

iSuperpose includes the following softwares, which are compiled upon installation

- TMalign (http://zhanglab.ccmb.med.umich.edu/TM-align/) used for protein structures alignment
- QBestFit (http://bioserv.rpbs.jussieu.fr/software/QBestFit.c) used to compute transformations associated to coordinates superimposition; this program computes the RMSd associated to superimposition
- GDT_TS is a fork of QBestFit to compute the GDT_TS score

In addition, it depends on the PyPDB class

NOTA BENE
---------

This software is fully functionnal, but please don't look at the source code. I started to write it when I was a Python's noob and I never took the time to refactorize the code. One day I will, and at this time I will upload it to PyPI.
""",

    cmdclass={
        'build': iSuperposeBuild,
        'install': iSuperposeInstall,
    }
)